/*
 *   Copyright (C) 2007 Matt Broadstone <mbroadst@kde.org>
 *   Copyright (C) 2007 Matias Costa <m.costacano@gmail.com>
 *   Copyright (C) 2007 Montel Laurent <montel@kde.org>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include <QDesktopWidget>
#include <QFileSystemWatcher>
#include <QApplication>
#include <QPainter>
#include <QGraphicsSceneMouseEvent>
#include <QGraphicsScene>
#include <QTimer>
#include <QAction>

#include <KDebug>
#include <KGlobalSettings>
#include <KAuthorized>
#include <KToggleAction>
#include <KFileItem>
#include <KDirLister>
#include <KIO/ListJob>
#include <KIO/Job>
#include <KIO/JobUiDelegate>
#include <KIO/UDSField>

#include "desktop.h"
#include "desktop.moc"
#include "launcher.h"
#include "plasmadesktopsettings.h"
#include "init.h"
#include "smartgroups.h"

inline QPointF readPos(const KConfigGroup &conf, const QString &key)
{
    QString stringPos = conf.readEntryUntranslated(key);
    QStringList coord = stringPos.split(",");
    kDebug() << stringPos<<coord[0].toDouble()<<coord[1].toDouble();
    return QPointF(coord[0].toDouble(), coord[1].toDouble());
}

inline void savePos(KConfigGroup &conf, const Launcher *launcher)
{
    QPointF pos = launcher->pos();
    conf.writeEntry(launcher->id(), QString("%1,%2").arg(pos.x()).arg(pos.y()));
}

inline void savePos(KConfigGroup &conf, const QList<Launcher*> &launchers)
{
    foreach(const Launcher *launcher, launchers) {
        savePos(conf, launcher);
    }
}


StandardDesktop::StandardDesktop(QObject *parent, const QVariantList &args)
    : Plasma::Applet(parent, args),
      m_dirLister(0), m_solidEngine(0), m_smartGroups(0)
{
    setFlag(QGraphicsItem::ItemIsMovable, false);
    PlasmaDesktopSettings::instance("plasma-desktop");
    testLocalInstallation();
    setZValue(0);

    // Solid and signals
    m_enableMedia = PlasmaDesktopSettings::mediaEnabled();
    
    // Organization
    m_verticalOrientation = true;
    // Smart placing disabled by default
    m_smartGroups = 0;
    
    //create menu
    createMenu();
    
    // Grid
    m_gridAlign = true;
    m_bLockIcon = !PlasmaDesktopSettings::lockIcons;
    setGridSize(QSizeF(100, 100));
}

StandardDesktop::~StandardDesktop()
{
    KConfigGroup config = globalConfig();
    savePos(config, m_desktopItems);
    delete m_solidEngine;
    delete m_dirLister;
}

void StandardDesktop::createMenu()
{
    QAction* alignHorizontal = new QAction(i18n("Line Up Horizontal"), this);
    actions.append(alignHorizontal);
    connect(alignHorizontal, SIGNAL(triggered(bool)), this , SLOT(slotAlignHorizontal()));
    QAction* alignVertical = new QAction(i18n("Line Up Vertical"), this);
    connect(alignVertical, SIGNAL(triggered(bool)), this , SLOT(slotAlignVertical()));
    actions.append(alignVertical);

    QAction* separator = new QAction(this);
    separator->setSeparator(true);
    actions.append(separator);
    m_lockIcon = new KToggleAction(i18n("Lock"),this);
    connect(m_lockIcon, SIGNAL(triggered(bool)), this , SLOT(slotLock(bool)));
    actions.append(m_lockIcon);
    m_alignToGrid = new KToggleAction(i18n("Align to grid"),this);
    connect(m_alignToGrid, SIGNAL(triggered(bool)), this , SLOT(slotAlignToGrid(bool)));
    actions.append(m_alignToGrid);
    m_groups = new KToggleAction(i18n("Smart Groups"),this);
    connect(m_groups, SIGNAL(triggered(bool)), this , SLOT(slotSmartGroups(bool)));
    actions.append(m_groups);
    
    separator = new QAction(this);
    separator->setSeparator(true);
    actions.append(separator);
    
    QAction* sortByNameSensitive = new QAction(i18n("Sort by Name (Case Sensitive)"), this);
    actions.append(sortByNameSensitive);
    connect(sortByNameSensitive, SIGNAL(triggered(bool)), this , SLOT(slotSortByNameCaseSensitive()));
    QAction* sortByNameInsensitive = new QAction(i18n("Sort by Name (Case Insensitive)"), this);
    actions.append(sortByNameInsensitive);
    connect(sortByNameInsensitive, SIGNAL(triggered(bool)), this , SLOT(slotSortByNameCaseInsensitive()));
    QAction* sortBySize = new QAction(i18n("Sort by Size"), this);
    actions.append(sortBySize);
    connect(sortBySize, SIGNAL(triggered(bool)), this , SLOT(slotSortBySize()));
    QAction* sortByType = new QAction(i18n("Sort by Type"), this);
    actions.append(sortByType);
    connect(sortByType, SIGNAL(triggered(bool)), this , SLOT(slotSortByType()));
}

QList<QAction*> StandardDesktop::contextActions()
{
  //Update action.
  m_alignToGrid->setChecked(m_gridAlign);
  m_lockIcon->setChecked(m_bLockIcon);
  m_groups->setChecked(m_smartGroups != 0);
  return actions;
}

void StandardDesktop::slotSortByNameCaseSensitive()
{
}

void StandardDesktop::slotSortByNameCaseInsensitive()
{
}

void StandardDesktop::slotSortBySize()
{
}

void StandardDesktop::slotSortByType()
{
}

void StandardDesktop::slotSortByDate()
{
}


void StandardDesktop::slotAlignToGrid(bool b)
{
    setGridAligned(b);
}

void StandardDesktop::slotSmartGroups(bool b){

    if (b && !m_smartGroups) {
        m_smartGroups = new SmartGroups(this);
    } else if (!b) {
        delete m_smartGroups;
        m_smartGroups = 0;
    }
}

void StandardDesktop::slotLock(bool lock)
{
   m_bLockIcon = !lock;
   foreach (Launcher *item, m_desktopItems)
        item->lockIcon(m_bLockIcon);
}

void StandardDesktop::slotAlignHorizontal()
{
    m_verticalOrientation = false;
    alignHorizontal(m_desktopItems);
    if (m_smartGroups) {
        m_smartGroups->reset();
    }
}

void StandardDesktop::slotAlignVertical()
{
    m_verticalOrientation = true;
    alignVertical(m_desktopItems);
    if (m_smartGroups) {
        m_smartGroups->reset();
    }
}

void StandardDesktop::changeAlignment(bool horizontal)
{
    m_verticalOrientation = !horizontal;
    if (m_verticalOrientation) {
        alignVertical(m_desktopItems);
    } else {
        alignHorizontal(m_desktopItems);
    }
}

void StandardDesktop::configureMedia()
{
    if(m_enableMedia) {
       if(!m_solidEngine) {
           m_solidEngine = dataEngine("solidnotifierengine");
           connect(m_solidEngine, SIGNAL(newSource(const QString&)),
                     this, SLOT(sourceAdded(const QString&)));
           //TODO update it.
       }
    } else if(m_solidEngine) {
        delete m_solidEngine;
        m_solidEngine=0;
        // remove from the lists and delete
        foreach(Launcher *launcher, m_solidDevices) {
            m_desktopItems.removeAll(launcher);
            launcher->deleteLater();
        }
        m_solidDevices.clear();
    }
}

void StandardDesktop::init()
{
    configureMedia();
    // Desktop files
    if (!m_dirLister) {
        m_dirLister = new KDirLister();
        connect(m_dirLister, SIGNAL(clear()), this, SLOT(clearDesktop()));
        connect(m_dirLister, SIGNAL(completed()), this, SLOT(listingCompleted()));
        connect(m_dirLister, SIGNAL(newItems(const KFileItemList&)),
                this, SLOT(newItemsFound(const KFileItemList&)));
        connect(m_dirLister, SIGNAL(deleteItem(KFileItem*)),
                this, SLOT(itemDeleted(KFileItem*)));

        KUrl desktopUrl(KGlobalSettings::desktopPath());
        KAuthorized::allowUrlAction("list", KUrl(), desktopUrl);
        m_dirLister->setShowingDotFiles(PlasmaDesktopSettings::showHidden);
        m_dirLister->openUrl(desktopUrl);
    }
}

bool StandardDesktop::isGridAligned() const
{
    return m_gridAlign;
}

void StandardDesktop::setGridAligned(bool align)
{
    // if icon alignment is false now and we change
    // to true then align all icons
    if (!m_gridAlign && align) {
        foreach(Launcher *l, m_desktopItems) {
            alignToGrid(l);
        }
    }
    m_gridAlign = align;
}

void StandardDesktop::alignToGrid(Launcher *item)
{
    QPointF currentPos = mapToParent(item->pos());
    QPoint currentGridPos = mapToGrid(currentPos);
    alignToGrid(item, currentGridPos);
}

void StandardDesktop::alignToGrid(Launcher *item, const QPoint &pos)
{
    qreal width = item->sizeHint().width();
    qreal height = item->sizeHint().height();
    QPointF scenePos = mapFromGrid(pos);
    scenePos.rx() -= width/2;
    scenePos.ry() -= height/2;
    item->setPos(scenePos);
}

QSizeF StandardDesktop::gridSize() const
{
    return m_gridSize;
}

QSize StandardDesktop::gridDimensions() const
{
    QSizeF thisSize = contentSizeHint();
    return QSize( int(thisSize.width() / m_gridSize.width()),
                  int(thisSize.height() / m_gridSize.height()) );
}

void StandardDesktop::setGridSize(const QSizeF& gridSize)
{
    QSizeF desktopSize = contentSizeHint();
    qreal bestMachWidth = desktopSize.width()
                        / qRound(desktopSize.width()/gridSize.width());
    qreal bestMachHeight = desktopSize.height()
                        / qRound(desktopSize.height()/gridSize.height());
    m_gridSize = QSizeF(bestMachWidth, bestMachHeight);
    if (m_gridAlign) {
        foreach(Launcher *item, m_desktopItems) {
            alignToGrid(item);
        }
    }
}

QSizeF StandardDesktop::contentSizeHint() const
{
    // This should probably use KWin, this is easy right now..
    // TODO fix for multihead
     return QSizeF(QApplication::desktop()->screenGeometry().size());
}

#ifdef PAINTINTERFACE
void StandardDesktop::paintInterface(QPainter *painter, const QStyleOptionGraphicsItem *option,
                            const QRect &rect)
{
   
    Q_UNUSED(option);
    Q_UNUSED(rect);
/*    painter->setPen(Qt::red);
    painter->drawRect(rect);*/
    
    if (!m_smartGroups) return;
    painter->save();
    QColor color;
    painter->setPen(Qt::black);
    int cCount = m_smartGroups->groupCount();
    painter->setRenderHint( QPainter::Antialiasing );
    painter->setPen(Qt::NoPen);
    painter->setOpacity(0.5);
    for (int i=0; i<cCount; ++i ) {
        const Group &c = (*m_smartGroups)[i];
        color.setHsv(i * 360/cCount, 180, 255 );
        painter->setBrush(color);
        int uCount = c.count();
        for (int j=0; j<uCount; ++j ) {
            const Launcher *unit = c[j];
            QRectF rect(unit->pos().x(),
                        unit->pos().y(),
                                  20, 20);
            painter->drawEllipse(rect);
        }
    }

/*    QSizeF desktopSize = contentSizeHint();
    for (qreal i=m_gridSize.width(); i<desktopSize.width(); i+=m_gridSize.width()) {
        painter->drawLine(QPointF(i, 0), QPointF(i, desktopSize.height()));
    }
    for (qreal i=m_gridSize.height(); i<desktopSize.height(); i+=m_gridSize.height()) {
        painter->drawLine(QPointF(0, i), QPointF(desktopSize.width(), i));
    }*/
    painter->restore();
}
#endif

void StandardDesktop::clearDesktop()
{
    foreach (Launcher *item, m_desktopItems)
        item->deleteLater();

    m_desktopItems.clear();
    m_fileItems.clear();
    m_solidDevices.clear();
    if (m_smartGroups) m_smartGroups->clear();

    // ^^^^ This is incomplete
}

void StandardDesktop::listingCompleted()
{
    // we may eventually want to organize on this call
}

void StandardDesktop::newItemsFound(const KFileItemList &items)
{
    QList<Launcher*> newLaunchers;
    foreach (KFileItem *item, items)
    {
        if ( (item->isDir() && (item->name() == "." || item->name() == "..")) ||
                item->isHidden())
            continue;

        if (m_fileItems.contains(item))    // Find a better way to hash this
            continue;

        FileLauncher *tmp = new FileLauncher(item, this);
        tmp->installSceneEventFilter(this);

        // kDebug() << "New launcher: " << tmp->size() << tmp->boundingRect() << endl;
        newLaunchers << tmp;
        m_desktopItems << tmp;
        m_fileItems[item] = tmp;

        // Place out of screen to not detect them as an ocuppied place
        // We will place it in a propper position later
        tmp->setPos(-100, -100);
        tmp->setZValue(zValue()+1);
    }
    organize(newLaunchers);
}

void StandardDesktop::itemDeleted(KFileItem *file)
{
    Launcher *item = m_fileItems.take(file);
    m_desktopItems.removeAll(item);
    if (m_smartGroups) m_smartGroups->remove(item);
    globalConfig().deleteEntry(item->id());
    item->deleteLater();
}

void StandardDesktop::updated(const QString &source, Plasma::DataEngine::Data data)
{
    Q_UNUSED(source);
    QString deviceUdi = data["udi"].toString();
    bool newDevice = data["added"].toBool();

    kDebug() << "new device: added[" << newDevice << "] udi[" << deviceUdi << "]";

    if (newDevice)
    {
        if (!m_solidDevices[deviceUdi])
        {
            DeviceLauncher *launcher = new DeviceLauncher(deviceUdi, this);
            launcher->installSceneEventFilter(this);

            launcher->setPos(100, 100);
        }
        else
        {
            kDebug() << "added twice?";
        }
    }
    else
    {
        DeviceLauncher *launcher = m_solidDevices[deviceUdi];
        if (launcher)
            launcher->deleteLater();
    }
}

void StandardDesktop::sourceAdded(const QString &source)
{
    kDebug() << "sourceAdded";
    m_solidEngine->connectSource(source, this);
}

bool StandardDesktop::sceneEventFilter(QGraphicsItem *target, QEvent *event)
{
    switch (event->type())
    {
        case QEvent::GraphicsSceneMouseRelease:
            {
                if (m_gridAlign)
                {
                    bool resetGroups = false;
                    foreach (QGraphicsItem *item, scene()->selectedItems())
                    {
                        Launcher *launcher = qgraphicsitem_cast<Launcher*>(item);
                        if (launcher) {
                            alignToGrid(launcher);
                            resetGroups = true;
                        }
                    }
                    if (resetGroups && m_smartGroups) m_smartGroups->reset();
                }
            }
            break;

        default:
            break;
    }

    return QGraphicsItem::sceneEventFilter(target, event);
}

void StandardDesktop::alignHorizontal(const QList<Launcher*> &items)
{
    Launcher *icon;
    int index = 0;
    int itemCount = items.count();
    if (itemCount == 0) {
        return;
    } else {
        icon = items[0];
    }

    qreal desktopWidth = contentSizeHint().width();
    qreal desktopHeigth = contentSizeHint().height();
    qreal gridWidth = gridSize().width();
    qreal gridHeight = gridSize().height();

    QPointF pos;
    for(pos.ry() = gridHeight/2; pos.y() < desktopHeigth; pos.ry() += gridHeight) {
        for(pos.rx() = gridWidth/2; pos.x() < desktopWidth; pos.rx() += gridWidth) {
            QGraphicsItem *existing = scene()->itemAt(mapToParent(pos));
            // If the existing index is in the items list with an index
            // greater than icon index current we will have to move it.
            // So ignore it
            // If its equal then existing == icon
            int existingIndex = items.indexOf((Launcher*)existing);
            if (!existing || (existing == this) || (existingIndex>=index)) {
                alignToGrid(icon, mapToGrid(pos));
                // get next
                if (++index < itemCount) {
                    icon = items[index];
                } else {
                    /* all items located */
                    return;
                }
            }
        }
    }
    // TODO What to do if we run out of space? make grid smaller?
}

void StandardDesktop::alignVertical(const QList<Launcher*> &items)
{
    Launcher *icon;
    int index = 0;
    int itemCount = items.count();
    if (itemCount == 0) {
        return;
    } else {
        icon = items[0];
    }

    qreal desktopWidth = contentSizeHint().width();
    qreal desktopHeigth = contentSizeHint().height();
    qreal gridWidth = gridSize().width();
    qreal gridHeight = gridSize().height();

    QPointF pos;
    for(pos.rx() = gridWidth/2; pos.x() < desktopWidth; pos.rx() += gridWidth) {
        for(pos.ry() = gridHeight/2; pos.y() < desktopHeigth; pos.ry() += gridHeight) {
            QGraphicsItem *existing = scene()->itemAt(mapToParent(pos));
            // If *existing is in items list with an index
            // greater than current icon index, then we will have to move
            // existing after.
            // So ignore it
            // If its equal then existing == icon then... we ignore it too
            int existingIndex = items.indexOf((Launcher*)existing);
            if (!existing || (existing == this) || (existingIndex>=index)) {
                alignToGrid(icon, mapToGrid(pos));
                // get next
                if (++index < itemCount) {
                    icon = items[index];
                } else {
                    /* all items located */
                    return;
                }
            }
        }
    }
    // TODO What to do if we run out of space? make grid smaller?
}



void StandardDesktop::organize(const QList<Launcher*> &items)
{
    KConfigGroup config = globalConfig();
    QList<Launcher*> unknowLaunchers;
    foreach(Launcher *launcher, items) {
        QString launcherId = launcher->id();
        if (config.hasKey(launcherId)) {
            QPointF pos = readPos(config, launcherId);
            kDebug() << launcher->id() <<"from pos";
            launcher->setPos(pos);
        } else {
            unknowLaunchers << launcher;
        }
    }
    // If unknowLaunchers is empty all icons have been placed from config
    if (unknowLaunchers.isEmpty()) {
        return;
    }
    
    if (m_smartGroups) {
        m_smartGroups->place(unknowLaunchers);
    } else {
        if(m_verticalOrientation) {
            alignVertical(unknowLaunchers);
        } else {
            alignHorizontal(unknowLaunchers);
        }
        if (m_smartGroups) {
            m_smartGroups->reset();
        }
    }    
}

QPointF StandardDesktop::findFreePlaceNear(QPointF point)
{
    // TODO make this faster
    QPointF p = mapFromGrid(mapToGrid(point));
    if (isFreePlace(p)) return p;
    
    int maxD = qMax(gridDimensions().width(),
                       gridDimensions().height());
    qreal gridWidth = gridSize().width();
    qreal gridHeight = gridSize().height();
    // d is the grid squares ahead to look from p
    // each iteration we try to look d grid squares from p
    // and store free places in freePlaces list
    QList<QPointF> freePlaces;
    for (int d=1; d<maxD && freePlaces.isEmpty(); ++d) {
        // These variables control we do not look out of screen
        qreal left = qMax(p.x() -d*gridWidth, gridWidth/2);
        qreal right = qMin(p.x() +d*gridWidth,
                           contentSizeHint().width() -gridWidth/2);
        qreal top = qMax(p.y() -d*gridHeight, gridHeight/2);
        qreal bottom = qMin(p.y() +d*gridHeight,
                            contentSizeHint().height() -gridHeight/2);
        QPointF pos;
        for (qreal x=left; x<=right; x += gridWidth) {
            pos = QPointF(x, top);
            if (isFreePlace(pos)) {
                freePlaces.append(pos);
            }
            pos = QPointF(x, bottom);
            if (isFreePlace(pos)) {
                freePlaces.append(pos);
            }

        }
        for (qreal y=top; y<=bottom; y += gridHeight) {
            pos = QPointF(left, y);
            if (isFreePlace(pos)) {
                freePlaces.append(pos);
            }
            pos = QPointF(right, y);
            if (isFreePlace(pos)) {
                freePlaces.append(pos);
            }
        }
    }
    double min = maxD * gridWidth * gridHeight;
    // Take from freePlaces the nearest
    // If freePlaces is empty, p is initialized in the first line as
    // mapFromGrid(mapToGrid(point));
    foreach(QPointF free, freePlaces) {
        double d = qAbs(point.x() - free.x()) + qAbs(point.y() -free.y());
        if (d < min) {
            min = d;
            p = free;
        }
    }
    return p;
}


bool StandardDesktop::isFreePlace(const QPointF &p) {
    QGraphicsItem *existing = scene()->itemAt(mapToParent(p));
    bool free = (!existing) || (existing == this);
//     kDebug() << p << "////" << existing <<"///"  <<(!existing) <<(existing == this) << free;
    return free;
}
        

K_EXPORT_PLASMA_APPLET(desktop, StandardDesktop)
