/*
 *   Copyright (C) 2007 Matt Broadstone <mbroadst@kde.org>
 *   Copyright (C) 2007 Matias Costa <m.costacano@gmail.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#ifndef STDDESKTOP_H
#define STDDESKTOP_H

#include <QHash>

#include <plasma/applet.h>
// #define PAINTINTERFACE

namespace Plasma
{
    class DataEngine;
}

class KDirLister;
class Launcher;
class FileLauncher;
class DeviceLauncher;
class KFileItemList;
class KFileItem;
class KToggleAction;
class SmartGroups;

class StandardDesktop : public Plasma::Applet
{
    Q_OBJECT
public:
    StandardDesktop(QObject *parent, const QVariantList &args);
    ~StandardDesktop();

    QSizeF contentSizeHint() const;
#ifdef PAINTINTERFACE
    void paintInterface(QPainter *painter, const QStyleOptionGraphicsItem *option,
               const QRect& rect);
#endif
    bool isGridAligned() const;
    void setGridAligned(bool align);

    QSizeF gridSize() const;

    // Temporarily public
    QPointF findFreePlaceNear(QPointF p);
    void organize(const QList<Launcher*> &items);
    void alignToGrid(Launcher *item);
    void alignToGrid(Launcher *item, const QPoint &pos);

    /**
     * Get the grid square index where a point in local coordinates is located.
     * In normal circustances desktop coordinates is screen position.
     * A grid place is like a square in a checkboard.
     *
     *
     * @param pos A point in desktop coordinates
     * @return The grid 2D index pos is located.
     */
    inline QPoint mapToGrid(const QPointF &pos) const;
    
    /**
     * A grid place is like a square in a checkboard, and this method returns
     * its center point. So if you want the top left corner gridSize()/2 must
     * be substracted.
     *
     * @param pos A grid place index
     * @return The center point of the grid square indexed by pos.
     */
    inline QPointF mapFromGrid(const QPoint &pos) const;

    /**
     * Sets the grid size to the nearest size which results in
     * a integer number of grid places. If the desktop width is 1024
     * and the desired grid width is 100 the resulting gridSize will be
     * 102.4 because it's the nearest width with a integer number of
     * divisions (10)
     *
     * @arg gridSize The desired grid size to be set.
     */
    void setGridSize(const QSizeF& gridSize);

    /**
     * The width and height of the size returned by this method represents
     * the grid places a Launcher can be placed. A grid place is like a
     * square in a checkboard, this size represents the white and black squares
     * for each border.
     *
     * @return The grid places for each dimension
     */
    QSize gridDimensions() const;

    void changeAlignment(bool horizontal);
    virtual QList<QAction*> contextActions();
    inline const QList<Launcher*>& desktopItems();
    
public Q_SLOTS:
    void updated(const QString &, QHash<QString, QVariant> );

private Q_SLOTS:
    void clearDesktop();
    void listingCompleted();
    void newItemsFound(const KFileItemList &items);
    void itemDeleted(KFileItem *file);

    void sourceAdded(const QString &source);
   
    void slotAlignHorizontal();
    void slotAlignVertical();
    void slotLock(bool);
    void slotAlignToGrid(bool b);
    void slotSmartGroups(bool b);

    void slotSortByNameCaseSensitive();
    void slotSortByNameCaseInsensitive();
    void slotSortBySize();
    void slotSortByType();
    void slotSortByDate();

    void configureMedia();
    void init();

protected:
    bool sceneEventFilter(QGraphicsItem *watched, QEvent *event);
//     QVariant itemChange(GraphicsItemChange change, const QVariant &value);
    void createMenu();

private:
    enum SortCriteria {NameCaseSensitive = 0, NameCaseInsensitive = 1, Size = 2, Type = 3, Date =4 };

    void alignVertical(const QList<Launcher*> &items);
    void alignHorizontal(const QList<Launcher*> &items);
    bool isFreePlace(const QPointF &p);
    
    KDirLister *m_dirLister;
    Plasma::DataEngine *m_solidEngine;

    QList<Launcher*> m_desktopItems;

    QHash<KFileItem*, FileLauncher*> m_fileItems;
    QHash<QString, DeviceLauncher*> m_solidDevices;
    // If m_smartGroups != 0 then smart placing is activated
    SmartGroups *m_smartGroups;

    QList<QAction*> actions;

    bool m_verticalOrientation;
    SortCriteria m_sortCriteria;
    bool m_gridAlign;
    bool m_bLockIcon;
    bool m_enableMedia;
    QSizeF m_gridSize;
    KToggleAction* m_lockIcon;
    KToggleAction* m_alignToGrid;
    KToggleAction* m_groups;
};

QPoint StandardDesktop::mapToGrid(const QPointF &pos) const
{
    return QPoint(int(pos.x()/m_gridSize.width()),
                  int(pos.y()/m_gridSize.height()));
}

QPointF StandardDesktop::mapFromGrid(const QPoint &pos) const
{
    return QPointF(pos.x()*m_gridSize.width() +m_gridSize.width()/2,
                  pos.y()*m_gridSize.height() +m_gridSize.height()/2);
}

inline const QList<Launcher*>& StandardDesktop::desktopItems()
{
    return m_desktopItems;
}

#endif

