/*
 *   Copyright (C) 2007 Matias Costa <m.costacano@gmail.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include "smartgroups.h"

#include <QStringList>
#include <QGraphicsScene>
#include <QSet>
#include <math.h>
#include "desktop.h"

/*
These are the p_arameters of the algorithm. Distances are normalized.
p_Wt:                   Controls the weight the tags resemblance has
                        in the distance value
p_validationThreshold:  The percentage a tag must appear in a grop to be valid
p_joinDistance:         If two groups are closer than this value and the union
                        is valid, then join then.
p_minDiameter:          Do not divide groups with a diameter less than this.
*/
const double p_Wt = 0.30;
const double p_validationThreshold = 0.70;
const double p_joinDistance = 0.25;
const double p_minDiameter = 0.20;

SmartGroups::SmartGroups(StandardDesktop *desktop): QObject(),
        m_groups(),
        m_desktop(desktop)
{
    reset();
}


void SmartGroups::clear()
{
    m_groups.clear();
}

void SmartGroups::reset()
{
    m_groups.clear();
    Group group;
    foreach(Launcher *l, m_desktop->desktopItems()) {
        group.add(l);
    }
    m_groups.append(group);
    learn();
}

SmartGroups::~SmartGroups()
{
}


void SmartGroups::learn()
{
    const Launcher *farUnits[2] = {0,0};
    bool hasDivided = false;
    int learnIterations = 0;
    do {
         hasDivided = false;
        /**
         * cCount changes in the iteration
         */
        int cCount = m_groups.count();
        for(int c = 0; c < cCount && c < m_groups.count(); ++c) {
            Group &group = m_groups[c];
            if (group.count() == 0) {
                continue;
            }
            if (group.isValid()) {
                continue;
            }
            if (validate(group)) {
                group.setValid(true);
                continue;
            }
            
            double diameter = 0;
            int count = group.count();
            /**
             * For all units calculate diameter distance
             * euclidean(a,b) == euclidean(b,a)
             */
            for(int i =0; i<count; ++i) {
                for(int j=i+1; j<count; ++j) {
                    double d = euclideanDistance(group[i],
                                                 group[j]);
                    if (d > diameter) {
                        /**
                         * Remember the diameter distance and
                         * the units who have it
                         */
                        diameter = d;
                        farUnits[0] = group[i];
                        farUnits[1] = group[j];
                    }
                }
            }
            if (diameter > p_minDiameter) {
                divide( c, farUnits[0], farUnits[1] );
                /**
                 * On dividing we remove the group in position c
                 * so next launcher is now in c and we have a
                 * launcher less in the list
                 */
                --c;
                --cCount;
                 hasDivided = true;
            } else {
                group.setValid(true);
            }
        }
        
        balance();
                
    } while (hasDivided && (++learnIterations)<20);
    join();
    balance();
    // Remove empty groups except one
    int cCount = m_groups.count();
    for(int c = 0; c < cCount; ++c) {
        Group &group = m_groups[c];
        if (group.count() == 0 && cCount>1) {
            m_groups.removeAt(c);
            --c;
            --cCount;
        }
    }
}

int SmartGroups::groupCount() const
{
    return m_groups.count();
}

/**
 * Return the group index which e belongs
 */
int SmartGroups::groupIndex(const Launcher *e) const
{

    int i = 0;
    foreach(const Group &group, m_groups){
        if (group.contains(e)) return i;
        ++i;
    }
    return -1;
}


/**
 * Divide the group with index c taking as refence units u1 and u2 removing it
 * Then groups are balanced looking resulting groups distance
 */
void SmartGroups::divide(int c, const Launcher *u1, const Launcher *u2)
{
    if (c >= groupCount()) {
//         qDebug() << "FATAL divide: no c!!!";
        return;
    }
    Group old = m_groups.takeAt(c);
    Group new1, new2;
    int count = old.count();
    /**
     * Divide looking reference launchers
     */
    for(int i =0; i<count; ++i) {
        const Launcher *current = old[i];
        double d1 = euclideanDistance(current, u1);
        double d2 = euclideanDistance(current, u2);
        if (d1 < d2) new1.add(current);
        else new2.add(current);
    }
    m_groups.append(new1);
    m_groups.append(new2);
}

/**
 * Return the group which e belongs
 */
const Group& SmartGroups::group(const Launcher *e) const
{
    foreach(const Group& group, m_groups){
        if (group.contains(e)) return group;
    }
    return m_groups[0];
}

void SmartGroups::balance()
{
    bool change = false;
    int cCount = m_groups.count();
    do {
        change = false;
        for (int source=0; source<cCount; ++source) {
            Group &c = m_groups[source];
            int units = c.count();
            for (int i=0; i<c.count(); ++i) {
                const Launcher *u = c[i];
                int best = bestGroup(u);
                if (source != best) {
                    c.remove(u);
                    c.setValid(false);
                    m_groups[best].add(u);
                    m_groups[best].setValid(false);
                    
                    change = true;
                    --units;
                }
            }
        }
    } while(change);
}

bool SmartGroups::validate(const Group &c) const
{
    foreach (QString tag, c.tags()) {
        double rate = c.tag(tag) / double(c.count());
        if (rate >= p_validationThreshold) return true;
    }
    return false;
}


/*!
    \fn SmartGroups::bestGroup(Launcher *u)
 */
int SmartGroups::bestGroup(const Launcher *u)
{
    int cCount = groupCount();
    double min = 10000;
    int best = 0;
    for (int c=0; c<cCount; ++c) {
        double d = groupDistance(m_groups[c], u, p_Wt);
        if ( d < min ) {
            min = d;
            best = c;
        }
    }
    return best;
}


/*!
    \fn SmartGroups::remove(Launcher *u)
 */
void SmartGroups::remove(const Launcher *u)
{
    int i = groupIndex(u);
    if (i != -1) {
        m_groups[i].remove(u);
        if (m_groups[i].count() == 0) {
            m_groups.removeAt(i);
        } 
    } else {
        kDebug() << "FATAL: removing not existing Launcher!!";
    }
}


/*!
    \fn SmartGroups::join()
 */
void SmartGroups::join()
{
    for(int i=0; i<m_groups.count(); ++i) {
        Group &c1 = m_groups[i];
        for(int j=i+1; j<m_groups.count(); ++j) {
            Group &c2 = m_groups[j];
            double d = euclideanDistance(c1.center(), c2.center());
            if (d < p_joinDistance ) {
                double units = c1.count() + c2.count();
                bool doJoin = false;
                // Look all tags
                foreach (const QString &tag, (c1.tags() + c2.tags())) {
                    double rate1 = (c1.tag(tag)+c2.tag(tag))/units;
                    if (rate1 > p_validationThreshold) {
                        doJoin = true;
                        break;
                    }
                }
                if (doJoin) {
                    c1.join(c2);
                    m_groups.removeAt(j);
                    --j;
                }
            }
        }
    }
}

void SmartGroups::place(const QList<Launcher*>& items)
{
    // TODO make this faster
//     m_deskto p->organize(items);
    foreach(Launcher *launcher, items) {
        double min = 10000;
        Group *best = 0;
        QString tags = launcher->tags();
        for(int i=0; i<m_groups.count(); ++i) {
            Group &c = m_groups[i];
            double d = tagsDistance(c, tags);
            foreach (QString tag, c.tags()) {
            }
            if (d < min) {
                min = d;
                best = &c;
            }
        }
        if (best) {
            QPointF pos = m_desktop->findFreePlaceNear(best->center());
            QPoint gridPlace = m_desktop->mapToGrid(pos);
            m_desktop->alignToGrid(launcher,gridPlace );
            best->add(launcher);
        } else {
            kDebug() << "Fatal: No best group!!";
        }
    }
//     reset();
}

double SmartGroups::euclideanDistance(const QPointF &a, const QPointF &b)
{
    QSizeF s = m_desktop->contentSizeHint();
    double dx = (a.x() - b.x())/(s.width());
    double dy = (a.y() - b.y())/(s.height());
    return  sqrtf((1.0/2.0)*dx*dx + (1.0/2.0)*dy*dy);
}

double SmartGroups::euclideanDistance(const Launcher *a, const Launcher *b)
{
    return euclideanDistance(a->pos(), b->pos());
}

double SmartGroups::tagsDistance(const Group &c,
                    const QString &tags)
{
    double rate = 0;
    double cCount = double(c.count());
    if (cCount == 0) return 1;
    QStringList tagList = tags.split(" ");
    foreach (QString tag, tagList) {
        rate += 1.0 - (c.tag(tag) / cCount);
    }
    return (1.0/tagList.count())*rate;
}

double SmartGroups::groupDistance(const Group &c,
                     const Launcher *e, double Wt)
{
    if (c.count() == 0) return 1;
    QSizeF s = m_desktop->contentSizeHint();
    double dx = (e->pos().x() - c.center().x())
                    /(s.width()-1);
    double dy = (e->pos().y() - c.center().y())
                    /(s.height()-1);
    double dt = tagsDistance(c, e->tags());
    return sqrtf((1.0/3.0)*dx*dx
            + (1.0/3.0)*dy*dy
            + (Wt/3.0)*dt*dt);
}
