/*
 *   Copyright (C) 2007 Petri Damsten <damu@iki.fi>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "news.h"
#include <QPainter>
#include <QSizeF>
#include <QGraphicsSceneDragDropEvent>
#include <krun.h>
#include <plasma/widgets/lineedit.h>
#include <plasma/widgets/boxlayout.h>

#define START "<html><head><meta name=\"qrichtext\" content=\"1\" />\
<style type=\"text/css\">\
.title { font-size:11pt; font-weight:normal; font-style:normal; color:#fcfcfc; \
background-color:#2f2f2f; } \
.date { font-size:9pt; font-weight:bold; font-style:normal; color:#fcfcfc; \
background-color:#2f2f2f; } \
td { font-size:9pt; font-weight:normal; font-style:normal; color:#fcfcfc; \
background-color:#1f1f1f; } \
table { background-color:#1f1f1f; } \
a { text-decoration:none; color:#fcfcfc } \
</style></head><body><table>"

#define TITLE "<tr><td colspan=\"2\" class=\"title\">\
<b><a href=\"%1\">%2</a></b></td></tr>"

#define DATE "<tr><td colspan=\"2\" class=\"date\">%2</td></tr>"

#define ITEM "<tr><td colspan=\"2\"><a href=\"%1\">&middot;&nbsp;%2</a></td></tr>"
#define ITEM_WITH_TIME "<tr><td>%1</td><td><a href=\"%2\">&middot;&nbsp;%3</a></td></tr>"

#define END "</table></body><html>"

News::News(QObject *parent, const QVariantList &args)
    : Plasma::Applet(parent, args), m_size(250,300), m_dialog(0)
{
    setHasConfigurationInterface(true);
    setDrawStandardBackground(true);

    m_newsEdit = new Plasma::LineEdit(this);
    m_newsEdit->setStyled(false);
    m_newsEdit->setCursor(Qt::ArrowCursor);
    m_newsEdit->setGeometry(QRectF(QPointF(0, 0), m_size));
    connect(m_newsEdit, SIGNAL(linkActivated(const QString&)),
            this, SLOT(linkActivated(const QString&)));
    m_newsEdit->setTextInteractionFlags(Qt::TextBrowserInteraction);

    KConfigGroup cg = config();
    m_interval = cg.readEntry("interval", 30);
    m_feed = cg.readEntry("feed", "http://www.kde.org/dotkdeorg.rdf");
    connectToEngine();
}

News::~News()
{
}

void News::setContentSize(const QSizeF& size)
{
    m_size = size;
    m_newsEdit->setGeometry(QRectF(QPointF(0, 0), m_size));
}

void News::connectToEngine()
{
    if (!m_feed.isEmpty()) {
        Plasma::DataEngine* engine = dataEngine("rss");
        engine->connectSource(m_feed, this, m_interval * 60 * 1000);
    }
}

QSizeF News::contentSizeHint() const
{
    return m_size;
}

void News::paintInterface(QPainter *p,
        const QStyleOptionGraphicsItem *option, const QRect &contentsRect)
{
    Q_UNUSED(p);
    Q_UNUSED(option);
    Q_UNUSED(contentsRect);
}

void News::showConfigurationInterface()
{
    if (m_dialog == 0) {
        m_dialog = new KDialog;
        m_dialog->setCaption(i18n("Configure News"));

        ui.setupUi(m_dialog->mainWidget());
        m_dialog->setButtons(KDialog::Ok | KDialog::Cancel | KDialog::Apply);
        connect(m_dialog, SIGNAL(applyClicked()), this, SLOT(configAccepted()));
        connect(m_dialog, SIGNAL(okClicked()), this, SLOT(configAccepted()));

        ui.intervalSpinBox->setValue(m_interval);
        ui.feedComboBox->setEditText(m_feed);

        KConfig feedsFile("data", QString("news/feeds"));
        m_defaultFeeds = feedsFile.group("feeds").entryMap();
        foreach (const QString& name, m_defaultFeeds.keys()) {
            ui.feedComboBox->addItem(name);
        }
    }
    m_dialog->show();
}

void News::configAccepted()
{
    m_interval = ui.intervalSpinBox->value();
    m_feed = ui.feedComboBox->currentText();

    KConfigGroup cg = config();
    if (m_defaultFeeds.keys().contains(m_feed)) {
        m_feed = m_defaultFeeds[m_feed];
    }
    cg.writeEntry("feed", m_feed);
    cg.writeEntry("interval", m_interval);

    connectToEngine();
}

void News::dropEvent(QGraphicsSceneDragDropEvent *event)
{
    kDebug();
    if (KUrl::List::canDecode(event->mimeData())) {
        KUrl::List urls = KUrl::List::fromMimeData(event->mimeData());

        if (urls.count() > 0) {
            event->accept();
            m_feed = urls.first().prettyUrl();
            m_interval = 30;
            connectToEngine();
        }
    }
}

void News::updated(const QString& source, const Plasma::DataEngine::Data &data)
{
    if (source == m_feed && !data.isEmpty()) {
        QDateTime previous = QDateTime::currentDateTime();
        QString html = START;
        html += QString(TITLE).arg(data["link"].toString())
                             .arg(data["title"].toString());
        QVariantList items = data["items"].toList();
        foreach (QVariant tmp, items) {
            QMap<QString, QVariant> item = tmp.toMap();
            uint timeStamp = item["time"].toUInt();
            QString save = html;

            QString title = item["title"].toString();
            title.replace( " ", "&nbsp;" );

            if (timeStamp != 0) {
                QDateTime time;
                time.setTime_t(timeStamp);
                if (previous.date() != time.date()) {
                    html += QString(DATE).arg(time.date().toString());
                }
                html += QString(ITEM_WITH_TIME).arg(time.time().toString("HH:mm"))
                                               .arg(item["link"].toString())
                                               .arg(title);
                previous = time;
            } else {
                html += QString(ITEM).arg(item["link"].toString())
                                     .arg(title);
            }
            // UGLY hack
            m_newsEdit->setHtml(html + END);
            if (m_newsEdit->heightForWidth(m_size.width()) > m_size.height()) {
                html = save;
                break;
            }
            // End of UGLY hack
        }
        html += END;
        m_newsEdit->setHtml(html);
        m_size = m_newsEdit->sizeHint();
        updateGeometry();
    }
}

void News::linkActivated(const QString &link)
{
    KRun::runUrl(KUrl(link), "text/html", 0);
}

#include "news.moc"
