/*
 *   Copyright (C) 2007 Petri Damsten <damu@iki.fi>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License as
 *   published by the Free Software Foundation; either version 2, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "svgwidget.h"
#include "plasma/svg.h"
#include <kdebug.h>
#include <QPainter>

namespace Plasma {

class SvgWidget::Private
{
public:
    Private() :
        minimum(0),
        maximum(100),
        value(0),
        format("%1"),
        mode(PlainSvg),
        image(0),
        sizeHint(QSizeF(0.0, 0.0)) {};

    void paint(QPainter *p, const QString& elementID)
    {
        QRectF elementRect = image->elementRect(elementID);
        image->paint(p, elementRect.topLeft(), elementID);
    }

    void paintBackground(QPainter *p)
    {
        paint(p, "background");
        p->save();
        // TODO: drawText
    }

    void paintForeground(QPainter *p)
    {
        p->restore();
        paint(p, "foreground");
    }

    int minimum;
    int maximum;
    int value;
    QString format;
    QString svg;
    Mode mode;
    Plasma::Svg *image;
    QSizeF sizeHint;
};

SvgWidget::SvgWidget(QGraphicsItem *parent) :
        Plasma::Widget(parent),
        d(new Private)
{
     kDebug();
}

SvgWidget::~SvgWidget()
{
}

void SvgWidget::setMaximum(int maximum)
{
    d->maximum = maximum;
}

int SvgWidget::maximum() const
{
    return d->maximum;
}

void SvgWidget::setMinimum(int minimum)
{
    d->minimum = minimum;
}

int SvgWidget::minimum() const
{
    return d->minimum;
}

void SvgWidget::setValue(int value)
{
    d->value = value;
    update();
}

int SvgWidget::value() const
{
    return d->value;
}

void SvgWidget::setFormat(const QString& format)
{
    d->format = format;
}

QString SvgWidget::format() const
{
    return d->format;
}

void SvgWidget::setSvg(QString svg)
{
    kDebug() << svg;
    d->svg = svg;
    delete d->image;
    d->image = new Plasma::Svg(svg);
    d->sizeHint = d->image->size();
}

QString SvgWidget::svg() const
{
    return d->svg;
}

void SvgWidget::setMode(Mode mode)
{
    d->mode = mode;
}

SvgWidget::Mode SvgWidget::mode() const
{
    return d->mode;
}

QSizeF SvgWidget::sizeHint() const
{
    return d->sizeHint;
}

void SvgWidget::paintWidget(QPainter *p,
                            const QStyleOptionGraphicsItem *option,
                            QWidget *widget)
{
    if (size() != d->image->size()) {
        kDebug() << "Resize:" << size();
        d->image->resize(size());
    }
    Q_UNUSED(option)
    Q_UNUSED(widget)
    QRectF rect(QPointF(0, 0), size());
    QRectF clipRect;
    qreal percentage = 0.0;

    if (d->maximum != d->minimum) {
        percentage = (qreal)d->value / (d->maximum - d->minimum);
    }
    p->setRenderHint(QPainter::SmoothPixmapTransform);

    switch (d->mode) {
    case PlainSvg:
        d->image->paint(p, rect);
        break;
    case BarMeter:
        d->paintBackground(p);

        clipRect = d->image->elementRect("bar");
        if (clipRect.width() > clipRect.height()) {
            clipRect.setWidth(clipRect.width() * percentage);
        } else {
            qreal bottom = clipRect.bottom();
            clipRect.setHeight(clipRect.height() * percentage);
            clipRect.moveBottom(bottom);
        }
        p->setClipRect(clipRect);
        d->paint(p, "bar");

        d->paintForeground(p);
        break;
    case AnalogMeter:
        d->paintBackground(p);

        // TODO Get these somehow from svg?
        int minrotate = 30;
        int maxrotate = 330;
        QPointF rotateCenter(rect.width() / 2, rect.height() / 2);

        qreal angle = percentage * (maxrotate - minrotate) + minrotate;

        p->translate(rotateCenter);
        p->rotate(angle);
        p->translate(-1 * rotateCenter);
        d->paint(p, "pointer");

        d->paintForeground(p);
        break;
    }
}

} // End of namepace

#include "svgwidget.moc"
